use std::collections::HashSet;

use itertools::Itertools;

#[derive(Clone, Copy, Debug)]
pub enum Material {
    Air,
    Lava,
    Steam,
}

pub fn parse_droplet(input: &String) -> Vec<Vec<Vec<Material>>> {
    let mut cubes: Vec<(usize, usize, usize)> = Vec::new();

    for line in input.lines() {
        if line.is_empty() {
            break;
        }

        let mut sp = line.split(',');
        cubes.push((
            sp.next().unwrap().parse().unwrap(),
            sp.next().unwrap().parse().unwrap(),
            sp.next().unwrap().parse().unwrap(),
        ));
    }

    let (x_min, x_max) = cubes.iter().map(|c| c.0).minmax().into_option().unwrap();
    let (y_min, y_max) = cubes.iter().map(|c| c.1).minmax().into_option().unwrap();
    let (z_min, z_max) = cubes.iter().map(|c| c.2).minmax().into_option().unwrap();
    let x_len = x_max - x_min + 3;
    let y_len = y_max - y_min + 3;
    let z_len = z_max - z_min + 3;

    let mut result = vec![vec![vec![Material::Air; z_len]; y_len]; x_len];
    for (x, y, z) in cubes {
        result[x - x_min + 1][y - y_min + 1][z - z_min + 1] = Material::Lava;
    }

    result
}

pub fn surface_area(droplet: &Vec<Vec<Vec<Material>>>) -> usize {
    let x_len = droplet.len();
    assert!(x_len > 0);
    let y_len = droplet[0].len();
    assert!(y_len > 0);
    let z_len = droplet[0][0].len();
    assert!(z_len > 0);

    let up = (0..x_len)
        .flat_map(|x| {
            (0..y_len)
                .map(move |y| (x, y, (0..z_len).collect()))
                .collect_vec()
        })
        .collect();
    let down = (0..x_len)
        .flat_map(|x| {
            (0..y_len)
                .map(move |y| (x, y, (0..z_len).rev().collect()))
                .collect_vec()
        })
        .collect();
    let left = (0..x_len)
        .flat_map(|x| {
            (0..z_len)
                .map(move |z| (x, z, (0..y_len).collect()))
                .collect_vec()
        })
        .collect();
    let right = (0..x_len)
        .flat_map(|x| {
            (0..z_len)
                .map(move |z| (x, z, (0..y_len).rev().collect()))
                .collect_vec()
        })
        .collect();
    let fore = (0..y_len)
        .flat_map(|y| {
            (0..z_len)
                .map(move |z| (y, z, (0..x_len).collect()))
                .collect_vec()
        })
        .collect();
    let back = (0..y_len)
        .flat_map(|y| {
            (0..z_len)
                .map(move |z| (y, z, (0..x_len).rev().collect()))
                .collect_vec()
        })
        .collect();

    let mut result = 0;
    result += surface_area_dir(droplet, &up, ArgOrder::XYZ);
    result += surface_area_dir(droplet, &down, ArgOrder::XYZ);
    result += surface_area_dir(droplet, &left, ArgOrder::XZY);
    result += surface_area_dir(droplet, &right, ArgOrder::XZY);
    result += surface_area_dir(droplet, &fore, ArgOrder::YZX);
    result += surface_area_dir(droplet, &back, ArgOrder::YZX);

    result
}

enum ArgOrder {
    XYZ,
    XZY,
    YZX,
}

fn surface_area_dir(
    droplet: &Vec<Vec<Vec<Material>>>,
    lines: &Vec<(usize, usize, Vec<usize>)>,
    order: ArgOrder,
) -> usize {
    let mut result = 0;
    for (a, b, c_range) in lines {
        let mut last_empty = true;
        for c in c_range {
            let occupied = match order {
                ArgOrder::XYZ => droplet[*a][*b][*c],
                ArgOrder::XZY => droplet[*a][*c][*b],
                ArgOrder::YZX => droplet[*c][*a][*b],
            };
            if last_empty && matches!(occupied, Material::Lava) {
                result += 1;
            }
            last_empty = matches!(occupied, Material::Air);
        }
    }
    result
}

pub fn outer_surface_area(droplet: &Vec<Vec<Vec<Material>>>) -> usize {
    let x_len = droplet.len();
    assert!(x_len > 0);
    let y_len = droplet[0].len();
    assert!(y_len > 0);
    let z_len = droplet[0][0].len();
    assert!(z_len > 0);

    let mut next = Vec::new();

    // top/bottom
    for x in 0..x_len {
        for y in 0..y_len {
            next.push((x, y, 0));
            next.push((x, y, z_len - 1));
        }
    }
    // left/right
    for x in 0..x_len {
        for z in 0..z_len {
            next.push((x, 0, z));
            next.push((x, y_len - 1, z));
        }
    }
    // fore/back
    for y in 0..y_len {
        for z in 0..z_len {
            next.push((0, y, z));
            next.push((x_len - 1, y, z));
        }
    }

    let mut visited: HashSet<_> = next.iter().map(|c| c.clone()).collect();
    let generators: [Box<dyn Fn(&Coord) -> Option<Coord>>; 6] = [
        Box::new(|c| go_fore(c)),
        Box::new(|c| go_back(c, x_len)),
        Box::new(|c| go_left(c)),
        Box::new(|c| go_right(c, y_len)),
        Box::new(|c| go_up(c)),
        Box::new(|c| go_down(c, z_len)),
    ];

    let mut result = 0;
    while let Some(n) = next.pop().clone() {
        for gen in &generators {
            if let Some(c) = gen(&n) {
                let occupied = droplet[c.0][c.1][c.2];
                if matches!(occupied, Material::Air) && visited.insert(c) {
                    next.push(c);
                } else if matches!(occupied, Material::Lava) {
                    result += 1;
                }
            }
        }
    }
    result
}

type Coord = (usize, usize, usize);

fn go_fore(c: &Coord) -> Option<Coord> {
    if c.0 > 0 {
        Some((c.0 - 1, c.1, c.2))
    } else {
        None
    }
}

fn go_back(c: &Coord, x_len: usize) -> Option<Coord> {
    if c.0 < x_len - 1 {
        Some((c.0 + 1, c.1, c.2))
    } else {
        None
    }
}

fn go_left(c: &Coord) -> Option<Coord> {
    if c.1 > 0 {
        Some((c.0, c.1 - 1, c.2))
    } else {
        None
    }
}

fn go_right(c: &Coord, y_len: usize) -> Option<Coord> {
    if c.1 < y_len - 1 {
        Some((c.0, c.1 + 1, c.2))
    } else {
        None
    }
}

fn go_up(c: &Coord) -> Option<Coord> {
    if c.2 > 0 {
        Some((c.0, c.1, c.2 - 1))
    } else {
        None
    }
}

fn go_down(c: &Coord, z_len: usize) -> Option<Coord> {
    if c.2 < z_len - 1 {
        Some((c.0, c.1, c.2 + 1))
    } else {
        None
    }
}
