use std::collections::HashMap;

use itertools::Itertools;

use crate::util::Dir;

#[derive(Debug)]
pub enum TurnDir {
    CW,
    CCW,
}

#[derive(Debug)]
pub enum PassInstr {
    Move(usize),
    Turn(TurnDir),
}

#[derive(Debug, Clone)]
pub struct Pose {
    pub pos: (usize, usize),
    pub orientation: Dir,
}

#[derive(Debug)]
pub enum MoveResult {
    Blocked,
    BlockedByWrapping,
    WrappedAround,
    Success,
}

impl MoveResult {
    pub fn was_blocked(&self) -> bool {
        matches!(self, MoveResult::Blocked | MoveResult::BlockedByWrapping)
    }

    pub fn wrapped_around(&self) -> bool {
        matches!(
            self,
            MoveResult::WrappedAround | MoveResult::BlockedByWrapping
        )
    }
}

pub trait Navigable {
    type Pose;
    type NavigationInstruction;

    fn forward(&self, from: &Self::Pose, steps: usize) -> (Self::Pose, MoveResult);
    fn exec_instruction(
        &self,
        from: &Self::Pose,
        inst: &Self::NavigationInstruction,
    ) -> (Self::Pose, MoveResult);
    fn initial_pose(&self) -> Self::Pose;
}

#[derive(Debug)]
pub struct SparseGrid {
    pub grid: Vec<Vec<Option<bool>>>,
    pub height: usize,
    pub width: usize,
    pub row_bounds: Vec<(usize, usize)>,
    pub col_bounds: Vec<(usize, usize)>,
}

impl SparseGrid {
    fn print(&self, pose: Option<&Pose>) {
        let check_pos = |y, x| {
            if let Some(pose) = pose {
                pose.pos.0 == y && pose.pos.1 == x
            } else {
                false
            }
        };

        for (y, row) in self.grid.iter().enumerate() {
            for (x, col) in row.iter().enumerate() {
                let c = if check_pos(y, x) {
                    assert!(*col == Some(false));
                    match pose.unwrap().orientation {
                        Dir::Right => '>',
                        Dir::Down => 'v',
                        Dir::Left => '<',
                        Dir::Up => '^',
                    }
                } else if col.is_none() {
                    ' '
                } else if *col == Some(true) {
                    '#'
                } else {
                    '.'
                };
                print!("{}", c);
            }
            print!("\n");
        }
    }
}

impl Navigable for SparseGrid {
    type Pose = Pose;
    type NavigationInstruction = PassInstr;

    fn forward(&self, from: &Pose, steps: usize) -> (Pose, MoveResult) {
        let mut pose = from.clone();
        let pos = &mut pose.pos;
        let mut status = MoveResult::Success;
        for _ in 0..steps {
            let next_pos = match from.orientation {
                Dir::Left => {
                    let bounds = &self.row_bounds[pos.0];
                    let next_col = if pos.1 == 0 || pos.1 == bounds.0 {
                        status = MoveResult::WrappedAround;
                        bounds.1
                    } else {
                        pos.1 - 1
                    };
                    (pos.0, next_col)
                }
                Dir::Up => {
                    let bounds = &self.col_bounds[pos.1];
                    let next_row = if pos.0 == 0 || pos.0 == bounds.0 {
                        status = MoveResult::WrappedAround;
                        bounds.1
                    } else {
                        pos.0 - 1
                    };
                    (next_row, pos.1)
                }
                Dir::Right => {
                    let bounds = &self.row_bounds[pos.0];
                    let next_col = if pos.1 == bounds.1 {
                        status = MoveResult::WrappedAround;
                        bounds.0
                    } else {
                        pos.1 + 1
                    };
                    (pos.0, next_col)
                }
                Dir::Down => {
                    let bounds = &self.col_bounds[pos.1];
                    let next_row = if pos.0 == bounds.1 {
                        status = MoveResult::WrappedAround;
                        bounds.0
                    } else {
                        pos.0 + 1
                    };
                    (next_row, pos.1)
                }
            };

            if self.grid[next_pos.0][next_pos.1].unwrap() {
                // We're about to hit a wall
                if status.wrapped_around() {
                    return (pose, MoveResult::BlockedByWrapping);
                } else {
                    return (pose, MoveResult::Blocked);
                }
            }
            *pos = next_pos;
        }

        (pose, status)
    }

    fn exec_instruction(
        &self,
        from: &Self::Pose,
        inst: &Self::NavigationInstruction,
    ) -> (Self::Pose, MoveResult) {
        match inst {
            PassInstr::Move(n) => self.forward(from, *n),
            PassInstr::Turn(dir) => (
                Pose {
                    pos: from.pos,
                    orientation: match dir {
                        TurnDir::CW => from.orientation.cw(),
                        TurnDir::CCW => from.orientation.ccw(),
                    },
                },
                MoveResult::Success,
            ),
        }
    }

    fn initial_pose(&self) -> Self::Pose {
        let col = self.grid[0]
            .iter()
            .position(|t| if let Some(occ) = t { !occ } else { false })
            .unwrap();
        Pose {
            pos: (0, col),
            orientation: Dir::Right,
        }
    }
}

#[derive(Debug, Hash, PartialEq, Eq, Clone, Copy)]
pub enum CubeSide {
    Top,
    Bottom,
    Left,
    Right,
    Fore,
    Back,
}

#[derive(Debug)]
pub struct CubeGrid {
    // The side grids don't have to be sparse, but we don't have a dense
    // implementation...
    pub sides: HashMap<CubeSide, SparseGrid>,

    pub side_height: usize,
    pub side_width: usize,
}

impl CubeGrid {
    pub fn from(sparse: &SparseGrid, pattern: &HashMap<CubeSide, (usize, usize)>) -> CubeGrid {
        let fold_width = pattern.values().map(|(_, col)| col).max().unwrap() + 1;
        let fold_height = pattern.values().map(|(row, _)| row).max().unwrap() + 1;
        let side_width = sparse.width / fold_width;
        assert!(sparse.width % fold_width == 0);
        let side_height = sparse.height / fold_height;
        assert!(sparse.height % fold_height == 0);

        let mut sides = HashMap::new();
        for (side, (fold_row, fold_col)) in pattern {
            let mut grid = Vec::new();
            for r in 0..side_height {
                let mut row = Vec::new();
                for c in 0..side_width {
                    let tile = sparse.grid[fold_row * side_height + r][fold_col * side_width + c];
                    assert!(tile.is_some());
                    row.push(tile);
                }
                grid.push(row);
            }
            let grid = SparseGrid {
                grid,
                height: side_height,
                width: side_width,
                row_bounds: vec![(0, side_width - 1); side_height],
                col_bounds: vec![(0, side_height - 1); side_width],
            };
            assert!(sides.insert(*side, grid).is_none());
        }

        CubeGrid {
            sides,
            side_height,
            side_width,
        }
    }

    pub fn print(&self, pose: &(CubeSide, Pose)) {
        for (side, grid) in &self.sides {
            if pose.0 != *side {
                continue;
            }
            println!("{:?}:", side);
            grid.print(Some(&pose.1));
            println!("");
        }
    }

    fn wrap_around(&self, side: CubeSide, pose: &Pose) -> (CubeSide, Pose) {
        // FIXME: This is hardcoded for the folding pattern of the actual input.
        // It won't work for the example input.
        assert_eq!(self.side_height, self.side_width);
        let dir = pose.orientation;
        let pos = pose.pos;
        let row = pos.0;
        let last = self.side_height - 1;
        let row_inv = last - row;
        let col = pos.1;
        match side {
            CubeSide::Top => match dir {
                Dir::Right => (
                    CubeSide::Right,
                    Pose {
                        pos: (row, 0),
                        orientation: Dir::Right,
                    },
                ),
                Dir::Down => (
                    CubeSide::Fore,
                    Pose {
                        pos: (0, col),
                        orientation: Dir::Down,
                    },
                ),
                Dir::Left => (
                    CubeSide::Left,
                    Pose {
                        pos: (row_inv, 0),
                        orientation: Dir::Right,
                    },
                ),
                Dir::Up => (
                    CubeSide::Back,
                    Pose {
                        pos: (col, 0),
                        orientation: Dir::Right,
                    },
                ),
            },
            CubeSide::Bottom => match dir {
                Dir::Right => (
                    CubeSide::Right,
                    Pose {
                        pos: (row_inv, last),
                        orientation: Dir::Left,
                    },
                ),
                Dir::Down => (
                    CubeSide::Back,
                    Pose {
                        pos: (col, last),
                        orientation: Dir::Left,
                    },
                ),
                Dir::Left => (
                    CubeSide::Left,
                    Pose {
                        pos: (row, last),
                        orientation: Dir::Left,
                    },
                ),
                Dir::Up => (
                    CubeSide::Fore,
                    Pose {
                        pos: (last, col),
                        orientation: Dir::Up,
                    },
                ),
            },
            CubeSide::Left => match dir {
                Dir::Right => (
                    CubeSide::Bottom,
                    Pose {
                        pos: (row, 0),
                        orientation: Dir::Right,
                    },
                ),
                Dir::Down => (
                    CubeSide::Back,
                    Pose {
                        pos: (0, col),
                        orientation: Dir::Down,
                    },
                ),
                Dir::Left => (
                    CubeSide::Top,
                    Pose {
                        pos: (row_inv, 0),
                        orientation: Dir::Right,
                    },
                ),
                Dir::Up => (
                    CubeSide::Fore,
                    Pose {
                        pos: (col, 0),
                        orientation: Dir::Right,
                    },
                ),
            },
            CubeSide::Right => match dir {
                Dir::Right => (
                    CubeSide::Bottom,
                    Pose {
                        pos: (row_inv, last),
                        orientation: Dir::Left,
                    },
                ),
                Dir::Down => (
                    CubeSide::Fore,
                    Pose {
                        pos: (col, last),
                        orientation: Dir::Left,
                    },
                ),
                Dir::Left => (
                    CubeSide::Top,
                    Pose {
                        pos: (row, last),
                        orientation: Dir::Left,
                    },
                ),
                Dir::Up => (
                    CubeSide::Back,
                    Pose {
                        pos: (last, col),
                        orientation: Dir::Up,
                    },
                ),
            },
            CubeSide::Fore => match dir {
                Dir::Right => (
                    CubeSide::Right,
                    Pose {
                        pos: (last, row),
                        orientation: Dir::Up,
                    },
                ),
                Dir::Down => (
                    CubeSide::Bottom,
                    Pose {
                        pos: (0, col),
                        orientation: Dir::Down,
                    },
                ),
                Dir::Left => (
                    CubeSide::Left,
                    Pose {
                        pos: (0, row),
                        orientation: Dir::Down,
                    },
                ),
                Dir::Up => (
                    CubeSide::Top,
                    Pose {
                        pos: (last, col),
                        orientation: Dir::Up,
                    },
                ),
            },
            CubeSide::Back => match dir {
                Dir::Right => (
                    CubeSide::Bottom,
                    Pose {
                        pos: (last, row),
                        orientation: Dir::Up,
                    },
                ),
                Dir::Down => (
                    CubeSide::Right,
                    Pose {
                        pos: (0, col),
                        orientation: Dir::Down,
                    },
                ),
                Dir::Left => (
                    CubeSide::Top,
                    Pose {
                        pos: (0, row),
                        orientation: Dir::Down,
                    },
                ),
                Dir::Up => (
                    CubeSide::Left,
                    Pose {
                        pos: (last, col),
                        orientation: Dir::Up,
                    },
                ),
            },
        }
    }
}

impl Navigable for CubeGrid {
    type Pose = (CubeSide, Pose);
    type NavigationInstruction = PassInstr;

    fn forward(
        &self,
        (from_side, from_pose): &Self::Pose,
        steps: usize,
    ) -> (Self::Pose, MoveResult) {
        let mut side = from_side.clone();
        let mut pose = from_pose.clone();
        let mut status = MoveResult::Success;
        for _ in 0..steps {
            let (next_pose, res) = self.sides[&side].forward(&pose, 1);
            if res.wrapped_around() {
                let (next_side, next_pose) = self.wrap_around(side, &pose);
                if self.sides[&next_side].grid[next_pose.pos.0][next_pose.pos.1] == Some(true) {
                    // Can't wrap around, we're blocked
                    return ((side, pose), MoveResult::BlockedByWrapping);
                }
                side = next_side;
                pose = next_pose;
                status = MoveResult::WrappedAround;
            } else if res.was_blocked() {
                return ((side, pose), MoveResult::Blocked);
            } else {
                pose = next_pose;
            }
        }

        ((side, pose), status)
    }

    fn exec_instruction(
        &self,
        from: &Self::Pose,
        inst: &Self::NavigationInstruction,
    ) -> (Self::Pose, MoveResult) {
        match inst {
            PassInstr::Move(n) => self.forward(from, *n),
            PassInstr::Turn(dir) => (
                (
                    from.0,
                    Pose {
                        pos: from.1.pos,
                        orientation: match dir {
                            TurnDir::CW => from.1.orientation.cw(),
                            TurnDir::CCW => from.1.orientation.ccw(),
                        },
                    },
                ),
                MoveResult::Success,
            ),
        }
    }

    fn initial_pose(&self) -> Self::Pose {
        let col = self.sides[&CubeSide::Top].grid[0]
            .iter()
            .position(|t| *t == Some(false))
            .unwrap();
        (
            CubeSide::Top,
            Pose {
                pos: (0, col),
                orientation: Dir::Right,
            },
        )
    }
}

pub fn parse_map_and_path(input: &str) -> (SparseGrid, Vec<PassInstr>) {
    let lines = input.lines().collect_vec();

    let mut grid: Vec<Vec<Option<bool>>> = Vec::new();
    let mut width = 0;
    for line in &lines[..lines.len() - 2] {
        width = width.max(line.len());
        let grid_line = line
            .chars()
            .map(|c| match c {
                ' ' => None,
                '.' => Some(false),
                '#' => Some(true),
                _ => panic!("Unknown grid character {}", c),
            })
            .collect();
        grid.push(grid_line);
    }

    let height = grid.len();
    let mut row_bounds = vec![(0, 0); height];
    for row in 0..height {
        let row_width = grid[row].len();
        if row_width < width {
            grid[row].extend((row_width..width).map(|_| None));
        }
        let lower = grid[row].iter().position(Option::is_some).unwrap();
        let upper = width - 1 - grid[row].iter().rev().position(Option::is_some).unwrap();
        row_bounds[row] = (lower, upper);
    }

    let mut col_bounds = vec![(0, 0); width];
    for col in 0..width {
        let col_it = (0..height).map(|row| grid[row][col]);
        let lower = col_it.clone().position(|t| t.is_some()).unwrap();
        let upper = height - 1 - col_it.rev().position(|t| t.is_some()).unwrap();
        col_bounds[col] = (lower, upper);
    }

    let grid = SparseGrid {
        grid,
        height,
        width,
        row_bounds,
        col_bounds,
    };

    let mut instructions = Vec::new();
    let mut inst_begin = 0;
    let inst_line = lines[lines.len() - 1].as_bytes();
    while inst_begin < inst_line.len() {
        let mut i = inst_begin;
        let c = inst_line[i];
        if c.is_ascii_digit() {
            while i < inst_line.len() && inst_line[i].is_ascii_digit() {
                i += 1;
            }
            let dist = std::str::from_utf8(&inst_line[inst_begin..i])
                .unwrap()
                .parse()
                .unwrap();
            instructions.push(PassInstr::Move(dist));
            inst_begin = i;
            continue;
        }

        let dir = match c {
            b'L' => TurnDir::CCW,
            b'R' => TurnDir::CW,
            _ => panic!("Unknown turn direction {}", c),
        };
        instructions.push(PassInstr::Turn(dir));
        inst_begin += 1;
    }

    (grid, instructions)
}
