use std::collections::{HashMap, HashSet};

use itertools::Itertools;

use crate::util::{self, SignedCoord};

pub fn parse_map(input: &str) -> HashSet<SignedCoord> {
    let mut result = HashSet::new();
    for (y, row) in input.lines().enumerate() {
        for (x, c) in row.chars().enumerate() {
            if c == '#' {
                result.insert((y as isize, x as isize));
            }
        }
    }
    result
}

const NUM_DIRS: usize = 4;

pub fn do_round(elves: &mut HashSet<SignedCoord>, it: usize) -> bool {
    let mut proposals: HashMap<SignedCoord, Vec<SignedCoord>> = HashMap::new();

    let mut elves_too_close = false;
    for elf in elves.iter() {
        if noone_adjacent(elves, elf) {
            continue;
        }
        elves_too_close = true;
        for offset in 0..NUM_DIRS {
            let dir = (it + offset) % NUM_DIRS;
            if let Some(new_loc) = propose(elves, elf, dir) {
                proposals.entry(new_loc).or_insert_with(Vec::new).push(*elf);
                break;
            }
        }
    }

    for (target, interested) in proposals {
        if interested.len() == 1 {
            elves.remove(&interested[0]);
            elves.insert(target);
        }
    }

    elves_too_close
}

fn noone_adjacent(elves: &HashSet<SignedCoord>, elf: &SignedCoord) -> bool {
    !(elves.contains(&(elf.0, elf.1 + 1))
        || elves.contains(&(elf.0, elf.1 - 1))
        || elves.contains(&(elf.0 + 1, elf.1))
        || elves.contains(&(elf.0 + 1, elf.1 + 1))
        || elves.contains(&(elf.0 + 1, elf.1 - 1))
        || elves.contains(&(elf.0 - 1, elf.1))
        || elves.contains(&(elf.0 - 1, elf.1 + 1))
        || elves.contains(&(elf.0 - 1, elf.1 - 1)))
}

fn propose(elves: &HashSet<SignedCoord>, elf: &SignedCoord, dir: usize) -> Option<SignedCoord> {
    let (moved, other) = match dir {
        0 => (
            (elf.0 - 1, elf.1),
            ((elf.0 - 1, elf.1 - 1), (elf.0 - 1, elf.1 + 1)),
        ), // N
        1 => (
            (elf.0 + 1, elf.1),
            ((elf.0 + 1, elf.1 - 1), (elf.0 + 1, elf.1 + 1)),
        ), // S
        2 => (
            (elf.0, elf.1 - 1),
            ((elf.0 - 1, elf.1 - 1), (elf.0 + 1, elf.1 - 1)),
        ), // E
        3 => (
            (elf.0, elf.1 + 1),
            ((elf.0 - 1, elf.1 + 1), (elf.0 + 1, elf.1 + 1)),
        ), // W
        _ => panic!("Unknown direction {}", dir),
    };
    if elves.contains(&moved) || elves.contains(&other.0) || elves.contains(&other.1) {
        None
    } else {
        Some(moved)
    }
}

pub fn print_map(elves: &HashSet<SignedCoord>) {
    let (y_min, y_max) = util::minmax_unwrap(&elves.iter().map(|c| c.0).minmax());
    let (x_min, x_max) = util::minmax_unwrap(&elves.iter().map(|c| c.1).minmax());

    println!("({}, {})", y_min, x_min);
    for y in y_min..(y_max + 1) {
        for x in x_min..(x_max + 1) {
            if elves.contains(&(y, x)) {
                print!("#");
            } else {
                print!(".");
            }
        }
        print!("\n");
    }
    print!("\n");
}

pub fn empty_ground(elves: &HashSet<SignedCoord>) -> usize {
    let (y_min, y_max) = util::minmax_unwrap(&elves.iter().map(|c| c.0).minmax());
    let (x_min, x_max) = util::minmax_unwrap(&elves.iter().map(|c| c.1).minmax());
    let y_dim = (y_max + 1 - y_min) as usize;
    let x_dim = (x_max + 1 - x_min) as usize;

    y_dim * x_dim - elves.len()
}
