use std::collections::HashSet;
use std::env;
use std::fs;
use std::hash::Hash;

use itertools::MinMaxResult;

pub fn parse_input() -> String {
    let args: Vec<String> = env::args().collect();
    let input_path = args
        .get(1)
        .expect(format!("Usage: {} INPUT-FILE", args.get(0).unwrap()).as_str());

    fs::read_to_string(input_path)
        .expect(format!("Couldn't read input file {}", input_path).as_str())
}

pub fn max_n<T: Ord + Copy>(slice: &[T], n: usize) -> Result<Vec<T>, ()> {
    if n == 0 || n > slice.len() {
        return Err(());
    }
    let mut max_vals = Vec::new();
    for _ in 0..n {
        max_vals.push(slice[0]);
    }

    for val in slice {
        if val > &max_vals[0] {
            max_vals[2] = max_vals[1];
            max_vals[1] = max_vals[0];
            max_vals[0] = *val;
        } else if val > &max_vals[1] {
            max_vals[2] = max_vals[1];
            max_vals[1] = *val;
        } else if val > &max_vals[2] {
            max_vals[2] = *val;
        }
    }

    Ok(max_vals)
}

pub type Coord = (usize, usize);

pub trait Coordinate {
    fn manhattan(&self, other: &Self) -> usize;
    fn to_signed(&self) -> SignedCoord;
    fn to_unsigned(&self) -> Coord;
}

impl Coordinate for Coord {
    fn manhattan(&self, other: &Self) -> usize {
        return self.0.abs_diff(other.0) + self.1.abs_diff(other.1);
    }

    fn to_signed(&self) -> SignedCoord {
        (self.0 as isize, self.1 as isize)
    }

    fn to_unsigned(&self) -> Coord {
        *self
    }
}

pub type SignedCoord = (isize, isize);

impl Coordinate for SignedCoord {
    fn manhattan(&self, other: &Self) -> usize {
        return self.0.abs_diff(other.0) + self.1.abs_diff(other.1);
    }

    fn to_signed(&self) -> SignedCoord {
        *self
    }

    fn to_unsigned(&self) -> Coord {
        (self.0 as usize, self.1 as usize)
    }
}

pub trait BnBState<T> {
    fn finished(&self) -> bool;
    fn lower_bound(&self, extra: &T) -> usize;
    fn upper_bound(&self, extra: &T) -> usize;
    fn possible_actions(&self, extra: &T) -> Vec<Self>
    where
        Self: Sized;
}

pub fn maximize<E, S>(initial_state: &S, extra: &E) -> S
where
    S: BnBState<E> + Clone + Hash + Eq,
{
    let mut lower_bound = initial_state.lower_bound(extra);
    let mut best = initial_state.clone();

    let mut next = vec![(initial_state.clone(), initial_state.upper_bound(extra))];
    let mut visited = HashSet::new();
    visited.insert(initial_state.clone());

    while let Some(n) = next.pop() {
        if n.1 < lower_bound {
            // Between pushing this state and popping it, we've found a better solution
            continue;
        }

        let state = n.0;
        for action in state.possible_actions(extra) {
            if action.finished() {
                let action_lower = action.lower_bound(extra);
                if action_lower > lower_bound {
                    lower_bound = action_lower;
                    best = action;
                }
            } else {
                let action_upper = action.upper_bound(extra);
                if action_upper > lower_bound && !visited.contains(&action) {
                    next.push((action.clone(), action_upper));
                    visited.insert(action);
                }
            }
        }
    }

    best
}

#[derive(Debug, Clone, Copy)]
pub enum Dir {
    Right = 0,
    Down,
    Left,
    Up,
}

impl Dir {
    pub fn from_usize(n: usize) -> Option<Dir> {
        match n {
            0 => Some(Dir::Right),
            1 => Some(Dir::Down),
            2 => Some(Dir::Left),
            3 => Some(Dir::Up),
            _ => None,
        }
    }

    pub fn from_char(c: char) -> Option<Dir> {
        match c {
            '^' => Some(Dir::Up),
            '>' => Some(Dir::Right),
            'v' => Some(Dir::Down),
            '<' => Some(Dir::Left),
            _ => None,
        }
    }

    pub fn to_str(&self) -> &'static str {
        match self {
            Dir::Right => ">",
            Dir::Down => "v",
            Dir::Left => "<",
            Dir::Up => "^",
        }
    }

    pub fn cw(&self) -> Dir {
        let d = *self as usize;
        Self::from_usize((d + 1) % 4).unwrap()
    }

    pub fn ccw(&self) -> Dir {
        let d = *self as usize;
        // Add 3 instead of subtracting one to avoid negative numbers
        Self::from_usize((d + 3) % 4).unwrap()
    }
}

pub fn minmax_unwrap<T: Clone>(mm: &MinMaxResult<T>) -> (T, T) {
    match mm {
        MinMaxResult::NoElements => panic!("No elements"),
        MinMaxResult::OneElement(a) => (a.clone(), a.clone()),
        MinMaxResult::MinMax(a, b) => (a.clone(), b.clone()),
    }
}
